@@ -11,6 +11,7 @@
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <netinet/in.h>
+#include <netinet/tcp.h>
 #include <arpa/inet.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -65,7 +66,7 @@ extern SSL_CTX  *ctx;
 #define K_SPECIAL 3
 #define K_FILE 4
 
-int 		list_transcripts( SNET *sn );
+int 		read_kfile( SNET *sn, char *kfile );
 
 int		f_quit( SNET *, int, char *[] );
 int		f_noop( SNET *, int, char *[] );
@@ -92,9 +93,10 @@ char		special_dir[ MAXPATHLEN ];
 char		command_file[ MAXPATHLEN ];
 char		upload_xscript[ MAXPATHLEN ];
 const EVP_MD    *md = NULL;
-struct list	*tran_list = NULL;
+struct list	*access_list = NULL;
 int		ncommands = 0;
 int		authorized = 0;
+int		prevstor = 0;
 char		hostname[ MAXHOSTNAMELEN ];
 
 extern int 	authlevel;
@@ -192,15 +194,17 @@ keyword( int ac, char *av[] )
     }
 
     if ( strcasecmp( av[ 1 ], "COMMAND" ) == 0 ) {
-	if ( ac != 2 ) {
+	if ( ac > 3 ) {
 	    return( -1 );
 	}
+	if ( ac == 2 ) {
+	    if ( strlen( command_file + 5 ) > MAXPATHLEN )  {
+		syslog( LOG_WARNING, "[tmp]/%s longer than MAXPATHLEN",
+			command_file );
+		return( -1 );
+	    }
+	}
 
-	if ( strlen( command_file + 5 ) > MAXPATHLEN )  {
-	    syslog( LOG_WARNING, "[tmp]/%s longer than MAXPATHLEN",
-		    command_file );
-	    return( -1 );
-	}
 	return( K_COMMAND );
     }
 
@@ -283,8 +287,37 @@ f_retr( SNET *sn, int ac, char **av )
 
     switch ( keyword( ac, av )) {
     case K_COMMAND:
-	/* command_file length checked in command_K */
-        strcpy( path, command_file );
+	if ( ac == 2 ) { 
+
+	    if ( snprintf( path, MAXPATHLEN, "command/%s", command_file )
+		    >= MAXPATHLEN ) {
+		syslog( LOG_ERR, "f_retr: command/%s: path too long",
+		    command_file );
+		snet_writef( sn, "%d Path too long\r\n", 540 );
+		return( 1 );
+	    }
+	} else {
+	    if (( d_path = decode( av[ 2 ] )) == NULL ) {
+		syslog( LOG_ERR, "f_retr: decode: buffer too small" );
+		snet_writef( sn, "%d Line too long\r\n", 540 );
+		return( 1 );
+	    } 
+
+	    /* Check for access */
+	    if ( !list_check( access_list, d_path )) {
+		syslog( LOG_WARNING | LOG_AUTH, "attempt to access: %s",
+		    d_path );
+		snet_writef( sn, "%d No access for %s\r\n", 540, d_path );
+		return( 1 );
+	    }
+
+	    if ( snprintf( path, MAXPATHLEN, "command/%s", d_path )
+		    >= MAXPATHLEN ) {
+		syslog( LOG_ERR, "f_retr: command path too long" );
+		snet_writef( sn, "%d Path too long\r\n", 540 );
+		return( 1 );
+	    }
+	}
 	break;
 
     case K_TRANSCRIPT:
@@ -295,7 +328,7 @@ f_retr( SNET *sn, int ac, char **av )
 	} 
 
 	/* Check for access */
-	if ( !list_check( tran_list, d_tran )) {
+	if ( !list_check( access_list, d_tran )) {
 	    syslog( LOG_WARNING | LOG_AUTH, "attempt to access: %s", d_tran );
 	    snet_writef( sn, "%d No access for %s\r\n", 540, d_tran );
 	    return( 1 );
@@ -342,7 +375,7 @@ f_retr( SNET *sn, int ac, char **av )
 	} 
 
 	/* Check for access */
-	if ( !list_check( tran_list, d_tran )) {
+	if ( !list_check( access_list, d_tran )) {
 	    syslog( LOG_WARNING | LOG_AUTH, "attempt to access: %s", d_tran );
 	    snet_writef( sn, "%d No access for %s:%s\r\n", 540, d_tran,
 		d_path );
@@ -363,7 +396,7 @@ f_retr( SNET *sn, int ac, char **av )
 	return( 1 );
     }
 
-    if ( ( fd = open( path, O_RDONLY, 0 ) ) < 0 ) {
+    if (( fd = open( path, O_RDONLY, 0 )) < 0 ) {
     	syslog( LOG_ERR, "open: %s: %m", path );
 	snet_writef( sn, "%d Unable to access %s.\r\n", 543, path );
 	return( 1 );
@@ -381,12 +414,11 @@ f_retr( SNET *sn, int ac, char **av )
 	return( 1 );
     }
 
-    snet_writef( sn, "%d Retrieving file\r\n", 240 );
     /*
      * Here's a problem.  Do we need to add long long support to
      * snet_writef?
      */
-    snet_writef( sn, "%" PRIofft "d\r\n", st.st_size );
+    snet_writef( sn, "240 Retrieving file\r\n%" PRIofft "d\r\n", st.st_size );
 
     /* dump file */
 
@@ -465,8 +497,36 @@ f_stat( SNET *sn, int ac, char *av[] )
 
     switch ( key = keyword( ac, av )) {
     case K_COMMAND:
-	/* command_file length checked in command_K */
-        strcpy( path, command_file );
+	if ( ac == 2 ) { 
+	    if ( snprintf( path, MAXPATHLEN, "command/%s", command_file )
+		    >= MAXPATHLEN ) {
+		syslog( LOG_ERR, "f_stat: command/%s: path too long",
+		    command_file );
+		snet_writef( sn, "%d Path too long\r\n", 540 );
+		return( 1 );
+	    }
+	} else {
+	    if (( d_path = decode( av[ 2 ] )) == NULL ) {
+		syslog( LOG_ERR, "f_stat: decode: buffer too small" );
+		snet_writef( sn, "%d Line too long\r\n", 540 );
+		return( 1 );
+	    } 
+
+	    /* Check for access */
+	    if ( !list_check( access_list, d_path )) {
+		syslog( LOG_WARNING | LOG_AUTH, "attempt to access: %s",
+		    d_path );
+		snet_writef( sn, "%d No access for %s\r\n", 540, d_path );
+		return( 1 );
+	    }
+
+	    if ( snprintf( path, MAXPATHLEN, "command/%s", d_path )
+		    >= MAXPATHLEN ) {
+		syslog( LOG_ERR, "f_stat: command path too long" );
+		snet_writef( sn, "%d Path too long\r\n", 540 );
+		return( 1 );
+	    }
+	}
 	break;
 
     case K_TRANSCRIPT:
@@ -477,7 +537,7 @@ f_stat( SNET *sn, int ac, char *av[] )
 	} 
 
 	/* Check for access */
-	if ( !list_check( tran_list, d_tran )) {
+	if ( !list_check( access_list, d_tran )) {
 	    syslog( LOG_WARNING | LOG_AUTH, "attempt to access: %s", d_tran );
 	    snet_writef( sn, "%d No access for %s\r\n", 540, d_tran );
 	    return( 1 );
@@ -535,10 +595,15 @@ f_stat( SNET *sn, int ac, char *av[] )
     snet_writef( sn, "%d Returning STAT information\r\n", 230 );
     switch ( key ) {
     case K_COMMAND:
-	snet_writef( sn, "%s %s %o %d %d %d %" PRIofft "d %s\r\n",
-		"f", "command", 
-		DEFAULT_MODE, DEFAULT_UID, DEFAULT_GID,
+	if ( ac == 2 ) {
+	    snet_writef( sn, "%s %s %o %d %d %d %" PRIofft "d %s\r\n",
+		"f", "command", DEFAULT_MODE, DEFAULT_UID, DEFAULT_GID,
 		st.st_mtime, st.st_size, cksum_b64 );
+	} else {
+	    snet_writef( sn, "%s %s %o %d %d %d %" PRIofft "d %s\r\n",
+		"f", av[ 2 ], DEFAULT_MODE, DEFAULT_UID, DEFAULT_GID,
+		st.st_mtime, st.st_size, cksum_b64 );
+	}
 	return( 0 );
         
 		    
@@ -620,9 +685,26 @@ f_stor( SNET *sn, int ac, char *av[] )
     char		*line;
     char		*d_tran, *d_path;
     int			fd;
+    int			zero = 0;
     off_t		len;
     ssize_t		rc;
     struct timeval	tv;
+    struct protoent	*proto;
+
+    if ( !prevstor ) {
+	/* Turn off TCP_NODELAY for stores */
+	if (( proto = getprotobyname( "tcp" )) == NULL ) {
+	    syslog( LOG_ERR, "f_stor: getprotobyname: %m" );
+	    return( -1 );
+	}
+
+	if ( setsockopt( snet_fd( sn ), proto->p_proto, TCP_NODELAY, &zero,
+		sizeof( zero )) != 0 ) {
+	    syslog( LOG_ERR, "f_stor: snet_setopt: %m" );
+	    return( -1 );
+	}
+	prevstor = 1;
+    }
 
     if ( checkuser && ( !authorized )) {
 	snet_writef( sn, "%d Not logged in\r\n", 551 );
@@ -824,7 +906,7 @@ f_starttls( SNET *sn, int ac, char **av )
      * Begin TLS
      */
     /* This is where the TLS start */
-    /* At this point the client is also staring TLS */
+    /* At this point the client is also starting TLS */
     /* 1 is for server, 0 is client */
     if (( rc = snet_starttls( sn, ctx, 1 )) != 1 ) {
         syslog( LOG_ERR, "f_starttls: snet_starttls: %s",
@@ -843,12 +925,12 @@ f_starttls( SNET *sn, int ac, char **av )
 	syslog( LOG_INFO, "CERT Subject: %s\n",
 	    X509_NAME_oneline( X509_get_subject_name( peer ), buf,
 	    sizeof( buf )));
-	X509_free( peer );
 
 	X509_NAME_get_text_by_NID( X509_get_subject_name( peer ),
 	    NID_commonName, buf, sizeof( buf ));
 	if (( remote_cn = strdup( buf )) == NULL ) {
 	    syslog( LOG_ERR, "strdup: %m" );
+	    X509_free( peer );
 	    return( -1 );
 	}
 	X509_free( peer );
@@ -864,7 +946,7 @@ f_starttls( SNET *sn, int ac, char **av )
 	commands  = auth;
 	ncommands = sizeof( auth ) / sizeof( auth[ 0 ] );
 
-	if ( list_transcripts( sn ) != 0 ) {
+	if ( read_kfile( sn, command_file ) != 0 ) {
 	    /* error message given in list_transcripts */
 	    exit( 1 );
 	}
@@ -1072,12 +1154,12 @@ command_k( char *path_config )
 	}
 
 	if (( remote_cn != NULL ) && wildcard( av[ 0 ], remote_cn, 0 )) {
-	    if ( snprintf( command_file, MAXPATHLEN, "command/%s", av[ 1 ] )
-		    >= MAXPATHLEN ) {
+	    if ( strlen( av[ 1 ] ) >= MAXPATHLEN ) {
 		syslog( LOG_ERR,
 		    "config file: line %d: command file too long\n", linenum );
 		continue;
 	    }
+	    strcpy( command_file, av[ 1 ] );
 	    if ( snprintf( temp, MAXPATHLEN, "%s/%s", special_dir,
 		    remote_cn ) >= MAXPATHLEN ) {
 		syslog( LOG_ERR, "config file: line %d: special dir too long\n",
@@ -1088,12 +1170,12 @@ command_k( char *path_config )
 	    return( 0 );
 	}
 	if ( wildcard( av[ 0 ], remote_host, 0 )) {
-	    if ( snprintf( command_file, MAXPATHLEN, "command/%s", av[ 1 ] )
-		    >= MAXPATHLEN ) {
+	    if ( strlen( av[ 1 ] ) >= MAXPATHLEN ) {
 		syslog( LOG_ERR,
 		    "config file: line %d: command file too long\n", linenum );
 		continue;
 	    }
+	    strcpy( command_file, av[ 1 ] );
 	    if ( snprintf( temp, MAXPATHLEN, "%s/%s", special_dir,
 		    remote_host ) >= MAXPATHLEN ) {
 		syslog( LOG_ERR, "config file: line %d: special dir too long\n",
@@ -1104,12 +1186,12 @@ command_k( char *path_config )
 	    return( 0 );
 	} 
 	if ( wildcard( av[ 0 ], remote_addr, 1 )) {
-	    if ( snprintf( command_file, MAXPATHLEN, "command/%s", av[ 1 ] )
-		    >= MAXPATHLEN ) {
+	    if ( strlen( av[ 1 ] ) >= MAXPATHLEN ) {
 		syslog( LOG_ERR,
 		    "config file: line %d: command file too long\n", linenum );
 		continue;
 	    }
+	    strcpy( command_file, av[ 1 ] );
 	    if ( snprintf( temp, MAXPATHLEN, "%s/%s", special_dir,
 		    remote_addr ) >= MAXPATHLEN ) {
 		syslog( LOG_ERR, "config file: line %d: special dir too long\n",
@@ -1129,65 +1211,125 @@ command_k( char *path_config )
 }
 
     int
-list_transcripts( SNET *sn )
+read_kfile( SNET *sn, char *kfile )
 {
-    int			ac, linenum;
-    char		kline[ MAXPATHLEN * 2 ];
-    FILE		*f;
-    char                **av;
+    int		ac;
+    int		linenum = 0;
+    char	**av;
+    char        line[ MAXPATHLEN ];
+    char	path[ MAXPATHLEN ];
+    ACAV	*acav;
+    FILE	*f;
 
-    if (( tran_list = list_new( )) == NULL ) {
+    if ( snprintf( path, MAXPATHLEN, "command/%s", kfile ) >= MAXPATHLEN ) {
 	snet_writef( sn,
 	    "%d Service not available, closing transmission channel\r\n", 421 );
-	syslog( LOG_ERR, "new_list: %m" );
+	syslog( LOG_ERR, "read_kfile: command/%s: path too long", kfile );
 	return( -1 );
     }
 
-    /* Create list of transcripts */
-    if (( f = fopen( command_file, "r" )) == NULL ) {
+    if (( acav = acav_alloc( )) == NULL ) {
 	snet_writef( sn,
 	    "%d Service not available, closing transmission channel\r\n", 421 );
-	syslog( LOG_ERR, "fopen: %s: %m", command_file );
+	syslog( LOG_ERR, "acav_alloc: %m" );
 	return( -1 );
     }
-    linenum = 0;
-    while ( fgets( kline, MAXPATHLEN, f ) != NULL ) {
+
+    if (( f = fopen( path, "r" )) == NULL ) {
+	snet_writef( sn,
+	    "%d Service not available, closing transmission channel\r\n", 421 );
+	syslog( LOG_ERR, "fopen: %s: %m", path );
+	return( -1 );
+    }
+
+    while ( fgets( line, MAXPATHLEN, f ) != NULL ) {
 	linenum++;
-	ac = acav_parse( NULL, kline, &av );
-	if (( ac == 0 ) || ( *av[ 0 ] == '#' )
-		|| ( *av[ 0 ] == 's')) {
+
+	ac = acav_parse( acav, line, &av );
+
+	if (( ac == 0 ) || ( *av[ 0 ] == '#' )) {
 	    continue;
 	}
+
 	if ( ac != 2 ) {
 	    snet_writef( sn,
 		"%d Service not available, closing transmission channel\r\n",
 		421 );
-	    syslog( LOG_ERR, "%s: %d: invalid command line\n",
-		    command_file, linenum );
-	    return( -1 );
+	    syslog( LOG_ERR, "%s: line %d: invalid number of arguments",
+		kfile, linenum );
+	    goto error;
 	}
-	if ( list_insert( tran_list, av[ 1 ] ) != 0 ) {
+
+	switch( *av[ 0 ] ) {
+	case 'k':
+	    if ( !list_check( access_list, av[ 1 ] )) {
+		if ( list_insert( access_list, av[ 1 ] ) != 0 ) {
+		    snet_writef( sn,
+	"%d Service not available, closing transmission channel\r\n", 421 );
+			syslog( LOG_ERR, "list_insert: %m" );
+		    goto error;
+		}
+		if ( read_kfile( sn, av[ 1 ] ) != 0 ) {
+		    goto error;
+		}
+	    }
+	    break;
+
+	case 'p':
+	case 'n':
+	    if ( !list_check( access_list, av[ 1 ] )) {
+		if ( list_insert( access_list, av[ 1 ] ) != 0 ) {
+		    snet_writef( sn,
+	"%d Service not available, closing transmission channel\r\n", 421 );
+			syslog( LOG_ERR, "list_insert: %m" );
+		    goto error;
+		}
+	    }
+	    break;
+
+	case 's':
+	    break;
+
+	default:
 	    snet_writef( sn,
 		"%d Service not available, closing transmission channel\r\n",
 		421 );
-	    syslog( LOG_ERR, "list_insert: %m" );
-	    return( -1 );
+	    syslog( LOG_ERR, "%s: line %d: %c: unknown file type", kfile,
+		linenum, *av[ 0 ] );
+	    goto error;
+
 	}
 
+	if ( ferror( f )) {
+	    snet_writef( sn,
+		"%d Service not available, closing transmission channel\r\n",
+		421 );
+	    syslog( LOG_ERR, "fgets: %m" );
+	    goto error;
+	}
     }
-    if ( ferror( f )) {
-	snet_writef( sn,
-	    "%d Service not available, closing transmission channel\r\n", 421 );
-	syslog( LOG_ERR, "fgets: %m" );
-	return( -1 );
-    }
-    if ( fclose( f ) < 0 ) {
+
+    if ( fclose( f ) != 0 ) {
 	snet_writef( sn,
 	    "%d Service not available, closing transmission channel\r\n", 421 );
 	syslog( LOG_ERR, "fclose: %m" );
+	goto error;
+    }
+
+    if ( acav_free( acav ) != 0 ) {
+	snet_writef( sn,
+	    "%d Service not available, closing transmission channel\r\n", 421 );
+	syslog( LOG_ERR, "acav_free: %m" );
 	return( -1 );
     }
+
     return( 0 );
+
+error:
+    fclose( f );
+    acav_free( acav );
+
+    return( -1 );
 }
 
     int
@@ -1197,6 +1339,7 @@ cmdloop( int fd, struct sockaddr_in *sin )
     struct hostent	*hp;
     char		*p;
     int			ac, i;
+    int			one = 1;
     unsigned int	n;
     char		**av, *line;
     struct timeval	tv;
@@ -1233,6 +1376,10 @@ cmdloop( int fd, struct sockaddr_in *sin )
     syslog( LOG_INFO, "child for [%s] %s",
 	    inet_ntoa( sin->sin_addr ), remote_host );
 
+    if ( setsockopt( fd, 6, TCP_NODELAY, &one, sizeof( one )) < 0 ) {
+	syslog( LOG_ERR, "setsockopt: %m" );
+    }
+
     if ( maxconnections != 0 ) {
 	if ( connections > maxconnections ) {
 	    snet_writef( sn, "%d Server busy\r\n", 420 );
@@ -1241,6 +1388,13 @@ cmdloop( int fd, struct sockaddr_in *sin )
 	    exit( 1 );
 	}
     }
+
+    if (( access_list = list_new( )) == NULL ) {
+	snet_writef( sn,
+	    "%d Service not available, closing transmission channel\r\n", 421 );
+	syslog( LOG_ERR, "new_list: %m" );
+	return( -1 );
+    }
     
     if ( authlevel == 0 ) {
 	/* lookup proper command file based on the hostname, IP or CN */
@@ -1250,8 +1404,8 @@ cmdloop( int fd, struct sockaddr_in *sin )
 		    remote_host );
 	    exit( 1 );
 	} else {
-	    if ( list_transcripts( sn ) != 0 ) {
-		/* error message given in list_transcripts */
+	    if ( read_kfile( sn, command_file ) != 0 ) {
+		/* error message given in read_kfile */
 		exit( 1 );
 	    }
 	    commands = auth;
@@ -1294,7 +1448,7 @@ cmdloop( int fd, struct sockaddr_in *sin )
 	    }
 	}
 	if ( i >= ncommands ) {
-	    snet_writef( sn, "%d Command %s unregcognized\r\n", 500, av[ 0 ] );
+	    snet_writef( sn, "%d Command %s unrecognized\r\n", 500, av[ 0 ] );
 	    continue;
 	}
 	if ( (*(commands[ i ].c_func))( sn, ac, av ) < 0 ) {
