@@ -25,10 +25,23 @@
 #include "cksum.h"
 #include "pathcmp.h"
 #include "largefile.h"
+#include "list.h"
 
-static struct transcript	*tran_head = NULL;
+int read_kfile( char *kfile, int location );
+
+struct transcript		*tran_head = NULL;
 static struct transcript	*prev_tran = NULL;
 extern int			edit_path;
+static int			foundspecial = 0;
+static char			*kdir;
+static struct list		*kfile_list;
+struct list			*special_list;
+
+char				*path_prefix = NULL;
+int				edit_path;
+int				skip;
+int				cksum;
+FILE				*outtran;
 
     void 
 transcript_parse( struct transcript *tran ) 
@@ -197,7 +210,7 @@ transcript_parse( struct transcript *tran )
     return;
 }
 
-    static void
+    void
 t_print( struct pathinfo *fs, struct transcript *tran, int flag ) 
 {
     struct pathinfo	*cur;
@@ -214,18 +227,31 @@ t_print( struct pathinfo *fs, struct transcript *tran, int flag )
 	cur = fs;	/* What if this is NULL? */
     }
 
+    /* Print name of transcript if it changed since the last t_print */
+    if (( edit_path == APPLICABLE )
+	    && (( flag == PR_TRAN_ONLY ) || ( flag == PR_DOWNLOAD )
+		|| ( flag == PR_STATUS_NEG ))
+	    && ( prev_tran != tran )) {
+	fprintf( outtran, "%s:\n", tran->t_shortname );
+	prev_tran = tran;
+    }
+
     /*
      * If a file is missing from the edit_path that was chosen, a - is 
      * printed and then the file name that is missing is printed.
      */
-    if (( edit_path == APPLICABLE ) && ( flag == PR_FS_ONLY )) {
-	fprintf( outtran, "- " );
-	cur = fs;
+    if ( edit_path == APPLICABLE ) {
+	if ( flag == PR_FS_ONLY ) {
+	    fprintf( outtran, "- " );
+	    cur = fs;
+	} else if ( flag == PR_STATUS_MINUS ) {
+	    fprintf( outtran, "- " );
+	}
     } else if (( edit_path ==  CREATABLE ) &&
 	    (( flag == PR_TRAN_ONLY ) || ( fs->pi_type == 'X' ))) {
 	fprintf( outtran, "- " );
 	cur = &tran->t_pinfo;
-    }
+    } 
 
     if (( epath = encode( cur->pi_name )) == NULL ) {
 	fprintf( stderr, "Filename too long: %s\n", cur->pi_name );
@@ -276,10 +302,6 @@ t_print( struct pathinfo *fs, struct transcript *tran, int flag )
     case 'f':
 	if (( edit_path == APPLICABLE ) && (( flag == PR_TRAN_ONLY ) || 
 		( flag == PR_DOWNLOAD ))) {
-	    if ( prev_tran != tran ) {
-		fprintf( outtran, "%s:\n", tran->t_shortname );
-		prev_tran = tran;
-	    }
 	    fprintf( outtran, "+ " );
 	}
 
@@ -538,14 +560,25 @@ transcript_select( void )
 	    }
 	}
 
-	if ( !begin_tran->t_pinfo.pi_minus ) {
-	    return( begin_tran );
+	/* This is presumably the NULL transcript. */
+	if ( !begin_tran->t_eof ) {
+	    /*
+	     * If the highest precedence transcript line has a leading '-',
+	     * then just pretend it's not there.
+	     */
+	    if ( begin_tran->t_pinfo.pi_minus ) {
+		transcript_parse( begin_tran );
+		continue;
+	    }
+
+	    /* Don't look outside of the initial path. */
+	    if ( !ischild( begin_tran->t_pinfo.pi_name, path_prefix )) {
+		transcript_parse( begin_tran );
+		continue;
+	    }
 	}
-	/*
-	 * If the highest precedence transcript line has a leading '-',
-	 * then just pretend it's not there.
-	 */
-	transcript_parse( begin_tran );
+
+	return( begin_tran );
     }
 }
 
@@ -634,7 +667,7 @@ transcript( char *path )
 }
 
     static void
-t_new( int type, char *fullname, char *shortname ) 
+t_new( int type, char *fullname, char *shortname, char *kfile ) 
 {
     struct transcript	 *new;
 
@@ -654,6 +687,7 @@ t_new( int type, char *fullname, char *shortname )
 	new->t_linenum = 0;
 	strcpy( new->t_shortname, shortname );
 	strcpy( new->t_fullname, fullname );
+	strcpy( new->t_kfile, kfile );
 	if (( new->t_in = fopen( fullname, "r" )) == NULL ) {
 	    perror( fullname );
 	    exit( 2 );
@@ -668,36 +702,22 @@ t_new( int type, char *fullname, char *shortname )
 }
 
     void
-transcript_init( char *kfile, int kfilemustexist )
+transcript_init( char *kfile, int location )
 {
-    char	**av;
     char	*special = "special.T";
-    char	line[ MAXPATHLEN ];
-    char	fullpath[ MAXPATHLEN ];
     char	*p;
-    int		length, ac, linenum = 0;
-    int		foundspecial = 0;
-    FILE	*fp;
-    char	*kdir;
+    char	fullpath[ MAXPATHLEN ];
 
     /*
      * Make sure that there's always a transcript to read, so other code
      * doesn't have to check it.
      */
-    t_new( T_NULL, NULL, NULL );
+    t_new( T_NULL, NULL, NULL, NULL );
 
     if ( skip ) {
 	return;
     }
 
-    if (( fp = fopen( kfile, "r" )) == NULL ) {
-	if ( kfilemustexist || ( edit_path == APPLICABLE )) {
-	    perror( kfile );
-	    exit( 2 );
-	}
-	return;
-    }
-
     if (( kdir = strdup( kfile )) == NULL ) {
         perror( "strdup failed" );
         exit( 2 );
@@ -709,13 +729,62 @@ transcript_init( char *kfile, int kfilemustexist )
         p++;
         *p = (char)'\0';
     }
+    if (( kfile_list = list_new( )) == NULL ) {
+	perror( "list_new" );
+	exit( 2 );
+    }
+    if ( list_insert( kfile_list, kfile ) != 0 ) {
+	perror( "list_insert" );
+	exit( 2 );
+    }
+    if (( special_list = list_new( )) == NULL ) {
+	perror( "list_new" );
+	exit( 2 );
+    }
+    if ( read_kfile( kfile, location ) != 0 ) {
+	exit( 2 );
+    }
+
+    if ( foundspecial && ( location == K_CLIENT )) {
+	/* open the special transcript if there were any special files */
+	if ( strlen( kdir ) + strlen( special ) + 2 > MAXPATHLEN ) {
+	    fprintf( stderr, 
+		    "special path too long: %s%s\n", kdir, special );
+	    exit( 2 );
+	}
+	sprintf( fullpath, "%s%s", kdir, special );
+	t_new( T_SPECIAL, fullpath, special, "special" );
+    }
+
+    if ( tran_head->t_type == T_NULL  && edit_path == APPLICABLE ) {
+	fprintf( stderr, "-A option requires a non-NULL transcript\n" );
+	exit( 2 );
+    }
+
+    return;
+}
+
+    int
+read_kfile( char *kfile, int location )
+{
+    int		length, ac, linenum = 0;
+    char	line[ MAXPATHLEN ];
+    char	fullpath[ MAXPATHLEN ];
+    char	*subpath;
+    char	**av;
+    FILE	*fp;
+
+    if (( fp = fopen( kfile, "r" )) == NULL ) {
+	perror( kfile );
+	return( -1 );
+    }
 
     while ( fgets( line, sizeof( line ), fp ) != NULL ) {
 	linenum++;
 	length = strlen( line );
 	if ( line[ length - 1 ] != '\n' ) {
 	    fprintf( stderr, "command: line %d: line too long\n", linenum );
-	    exit( 2 );
+	    return( -1 );
 	}
 
 	/* skips blank lines and comments */
@@ -726,54 +795,91 @@ transcript_init( char *kfile, int kfilemustexist )
 	if ( ac != 2 ) {
 	    fprintf( stderr, "command: line %d: expected 2 arguments, got %d\n",
 		    linenum, ac );
-	    exit( 2 );
+	    return( -1 );
 	} 
 
-	if ( strlen( kdir ) + strlen( av[ 1 ] ) + 2 > MAXPATHLEN ) {
-	    fprintf( stderr, "command: line %d: transcript name too long\n",
-		    linenum );
-	    fprintf( stderr, "command: line %d: %s%s\n",
-		    linenum, kdir, av[ 1 ] );
-	    exit( 2 );
+	switch( location ) {
+	case K_CLIENT:
+	    if ( snprintf( fullpath, MAXPATHLEN, "%s%s", kdir,
+		    av[ 1 ] ) >= MAXPATHLEN ) {
+		fprintf( stderr, "command: line %d: path too long\n",
+			linenum );
+		fprintf( stderr, "command: line %d: %s%s\n",
+			linenum, kdir, av[ 1 ] );
+		return( -1 );
+	    }
+	    break;
+
+	case K_SERVER:
+	    if ( *av[ 0 ] == 'k' ) {
+		subpath = "command";
+	    } else {
+		subpath = "transcript";
+	    }
+	    if ( snprintf( fullpath, MAXPATHLEN, "%s/%s/%s", _RADMIND_PATH,
+			subpath, av[ 1 ] ) >= MAXPATHLEN ) {
+		fprintf( stderr, "command: line %d: path too long\n",
+			linenum );
+		fprintf( stderr, "command: line %d: %s%s\n",
+			linenum, kdir, av[ 1 ] );
+		return( -1 );
+	    }
+	    break;
+
+	default:
+	    fprintf( stderr, "unknown location\n" );
+	    return( -1 );
 	}
-	sprintf( fullpath, "%s%s", kdir, av[ 1 ] );
 
 	switch( *av[ 0 ] ) {
-	case 'p':				/* positive */
-	    t_new( T_POSITIVE, fullpath, av[ 1 ] );
+	case 'k':				/* command file */
+	    if ( list_check( kfile_list, fullpath )) {
+		fprintf( stderr,
+		    "%s: line %d: command file loop: %s already included\n",
+		    kfile, linenum, av[ 1 ] );
+		return( -1 );
+	    }
+	    if ( list_insert( kfile_list, fullpath ) != 0 ) {
+		perror( "list_insert" );
+		return( -1 );
+	    }
+
+	    if ( read_kfile( fullpath, location ) != 0 ) {
+		return( -1 );
+	    }
 	    break;
+
 	case 'n':				/* negative */
-	    t_new( T_NEGATIVE, fullpath, av[ 1 ] );
+	    t_new( T_NEGATIVE, fullpath, av[ 1 ], kfile );
 	    break;
+
+	case 'p':				/* positive */
+	    t_new( T_POSITIVE, fullpath, av[ 1 ], kfile );
+	    break;
+
 	case 's':				/* special */
 	    foundspecial++;
+	    if ( location == K_SERVER ) {
+		if ( list_insert( special_list, av[ 1 ] ) != 0 ) {
+		    perror( "list_insert" );
+		    return( -1 );
+		}
+	    }
 	    continue;
+
 	default:
 	    fprintf( stderr, "command: line %d: '%s' invalid\n",
 		    linenum, av[ 0 ] );
-	    exit( 2 );
+	    return( -1 );
 	}
     }
 
-    fclose( fp );
-
-    /* open the special transcript if there were any special files */
-    if ( foundspecial ) {
-	if ( strlen( kdir ) + strlen( special ) + 2 > MAXPATHLEN ) {
-	    fprintf( stderr, 
-		    "special path too long: %s%s\n", kdir, special );
-	    exit( 2 );
-	}
-	sprintf( fullpath, "%s%s", kdir, special );
-	t_new( T_SPECIAL, fullpath, special );
+    if ( fclose( fp ) != 0 ) {
+	perror( kfile );
+	return( -1 );
     }
 
-    if ( tran_head->t_type == T_NULL  && edit_path == APPLICABLE ) {
-	fprintf( stderr, "-A option requires a non-NULL transcript\n" );
-	exit( 2 );
-    }
-
-    return;
+    return( 0 );
 }
 
     void
