@@ -24,11 +24,16 @@
 #define PR_DOWNLOAD	3 
 #define PR_STATUS	4 
 #define PR_STATUS_NEG	5
+#define PR_STATUS_MINUS	6
 
-int			edit_path;
-int			skip;
-int			cksum;
-FILE			*outtran;
+#define K_CLIENT	0
+#define K_SERVER	1
+
+extern int		edit_path;
+extern int		skip;
+extern int		cksum;
+extern FILE		*outtran;
+extern char		*path_prefix;
 
 struct pathinfo {
     char			pi_type;
@@ -46,16 +51,18 @@ struct transcript {
     int 		t_type;
     char		t_fullname[ MAXPATHLEN ];
     char		t_shortname[ MAXPATHLEN ];
+    char		t_kfile[ MAXPATHLEN ];
     int			t_linenum;
     int			t_eof;
     FILE		*t_in;
 };
 
 int			transcript( char * );
-void			transcript_init( char *, int );
+void			transcript_init( char *kfile, int location );
 struct transcript	*transcript_select( void );
 void			transcript_parse( struct transcript * );
 void			transcript_free( void );
+void			t_print( struct pathinfo *, struct transcript *, int );
 char			*hardlink( struct pathinfo * );
 int			hardlink_changed( struct pathinfo *, int );
 void			hardlink_free( void );
